#include "queue.h"

#include <iostream>
#include <vector>
#include <thread>
#include <mutex>
#include <cassert>
#include <array>
#include <atomic>

constexpr int COUNT = 100;
constexpr int WORKERS = 2;

std::array< std::atomic< int >, WORKERS * COUNT > seen;

void consume( Queue< int >& q ) {
    std::array< int, WORKERS > last;
    for ( auto &l : last )
        l = -1;

    for( int i = 0; i != COUNT; i++ ) {
        int x = q.pop();
        ++seen[ x ];
        int p = x / COUNT;
        assert( last[ p ] < x );
        last[ p ] = x;
    }
}

void produce( int initial, int count, Queue< int >& q ) {
    for ( int i = 0; i != count; i++ ) {
        q.push( initial + i );
    }
}

int main() {
    Queue< int > q;

    std::vector< std::thread > consumers;
    for ( int i = 0; i != WORKERS; i++ )
        consumers.emplace_back( consume, std::ref( q ) );

    std::vector< std::thread > producers;
    for ( int i = 0; i != WORKERS; i++ )
        producers.emplace_back( produce, i * COUNT, COUNT, std::ref( q ) );
        // try changing the count argument to COUNT - 1

    for ( auto& t: producers )
        t.join();

    // bonus: try running q.kill();

    for ( auto& t : consumers )
        t.join();

    for ( auto &v : seen )
        assert( v <= 1 );
}
